use crate::algorithm::Algorithm;
use crate::game::Game;

pub struct PrincipalVariationSearch {}

impl Algorithm for PrincipalVariationSearch {
    /// Principal Variation Search
    ///
    /// A more efficient alternative to Alpha-Beta Pruning
    /// https://en.wikipedia.org/wiki/Principal_variation_search
    fn compute<T: Game>(game: &T, depth: i32) -> Option<T::Move> {
        let mut cl = game.clone();
        let mut alpha = T::SCORE_MIN;
        let beta = T::SCORE_MAX;
        if depth == 0 || game.is_game_over() {
            None
        } else {
            let mut best_move = None;
            for (i, &mov) in game.possible_moves().iter().enumerate() {
                let undo = cl.do_move(mov);
                let mut score;
                if i == 0 {
                    score = -Self::pvs_rec(&mut cl, depth - 1, -beta, -alpha);
                } else {
                    score = -Self::pvs_rec(&mut cl, depth - 1, -alpha - 1, -alpha);
                    if alpha < score && score < beta {
                        score = -Self::pvs_rec(&mut cl, depth - 1, -beta, -score);
                    }
                }
                cl.undo_move(undo);
                if score > alpha {
                    best_move = Some(mov);
                    alpha = score;
                }
                if alpha >= beta {
                    break;
                }
            }
            best_move
        }
    }
}

impl PrincipalVariationSearch {
    fn pvs_rec<T: Game>(game: &mut T, depth: i32, mut alpha: i32, beta: i32) -> i32 {
        //println!("{:?}", game);
        if depth == 0 || game.is_game_over() {
            return game.get_score();
        }
        for (i, &mov) in game.possible_moves().iter().enumerate() {
            let undo = game.do_move(mov);
            let mut score;
            if i == 0 {
                score = -Self::pvs_rec(game, depth - 1, -beta, -alpha);
            } else {
                score = -Self::pvs_rec(game, depth - 1, -alpha - 1, -alpha);
                if alpha < score && score < beta {
                    score = -Self::pvs_rec(game, depth - 1, -beta, -score);
                }
            }
            game.undo_move(undo);
            if score > alpha {
                alpha = score;
            }
            if alpha >= beta {
                break;
            }
        }
        alpha
    }
}
