#![forbid(unsafe_code)]

use fox_and_hounds::FoxAndHounds;
use fox_and_hounds::Game;
use fox_and_hounds::NoughtsAndCrosses;
use fox_and_hounds::PrincipalVariationSearch;

use std::fmt;
use std::io;
use std::io::Write;
use std::process::exit;

enum ChosenGame {
    FoxAndHounds(FoxAndHounds),
    NoughtsAndCrosses(NoughtsAndCrosses),
}

impl fmt::Display for ChosenGame {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            ChosenGame::FoxAndHounds(x) => x.fmt(f),
            ChosenGame::NoughtsAndCrosses(x) => x.fmt(f),
        }
    }
}

/// Ask the user to choose a game
fn choose_game() -> ChosenGame {
    loop {
        println!("Which game would you like to play?");
        println!("F: Fox & Hounds");
        println!("N: Noughts and Crosses");
        print!("F/N: ");
        let _ = io::stdout().flush();

        let mut line = String::new();
        if io::stdin().read_line(&mut line).is_err() {
            println!("Failed to read input from stdin.");
            exit(1);
        }

        println!();

        line.make_ascii_lowercase();

        return match line.as_str().trim() {
            "f" => ChosenGame::FoxAndHounds(FoxAndHounds::default()),
            "n" => ChosenGame::NoughtsAndCrosses(NoughtsAndCrosses::default()),
            _ => {
                println!("Invalid input, please try again.");
                println!();
                continue;
            }
        };
    }
}

#[derive(Debug)]
enum GameType {
    PlayerFirst,
    PlayerSecond,
    TwoPlayers,
    TwoAIs,
}

/// Ask the user to choose a game type
fn choose_game_type() -> GameType {
    loop {
        println!("Which game type would you like to use?");
        println!("1: Player vs AI (default)");
        println!("2: AI vs Player");
        println!("3: Player vs Player");
        println!("4: AI vs AI");
        print!("[1 to 4]: ");
        let _ = io::stdout().flush();

        let mut line = String::new();
        if io::stdin().read_line(&mut line).is_err() {
            println!("Failed to read input from stdin.");
            exit(1);
        }

        println!();

        return match line.as_str().trim() {
            "1" | "" => GameType::PlayerFirst,
            "2" => GameType::PlayerSecond,
            "3" => GameType::TwoPlayers,
            "4" => GameType::TwoAIs,
            _ => {
                println!("Invalid input, please try again.");
                println!();
                continue;
            }
        };
    }
}

#[derive(Debug)]
enum MoveType {
    PlayerMove,
    AIMove,
}

fn move_type_for_move(game_type: &GameType, move_no: i32) -> MoveType {
    match (game_type, move_no % 2 == 0) {
        (GameType::PlayerFirst, true) => MoveType::PlayerMove,
        (GameType::PlayerFirst, false) => MoveType::AIMove,
        (GameType::PlayerSecond, true) => MoveType::AIMove,
        (GameType::PlayerSecond, false) => MoveType::PlayerMove,
        (GameType::TwoPlayers, _) => MoveType::PlayerMove,
        (GameType::TwoAIs, _) => MoveType::AIMove,
    }
}

fn do_ai_move(chosen_game: &ChosenGame) -> Option<ChosenGame> {
    match chosen_game {
        ChosenGame::FoxAndHounds(game) => {
            let mov = game.calculate_move::<PrincipalVariationSearch>(15, false)?;
            Some(ChosenGame::FoxAndHounds(game.execute_move(mov)))
        }
        ChosenGame::NoughtsAndCrosses(game) => {
            let mov = game.calculate_move::<PrincipalVariationSearch>(20, false)?;
            Some(ChosenGame::NoughtsAndCrosses(
                game.execute_move(mov),
            ))
        }
    }
}

fn get_possible_moves(chosen_game: &ChosenGame) -> Vec<String> {
    match chosen_game {
        ChosenGame::FoxAndHounds(game) => game
            .possible_moves()
            .into_iter()
            .map(|m| m.to_string())
            .collect(),
        ChosenGame::NoughtsAndCrosses(game) => game
            .possible_moves()
            .into_iter()
            .map(|m| m.to_string())
            .collect(),
    }
}

fn do_human_move(chosen_game: &ChosenGame, move_no: i32) -> Option<ChosenGame> {
    loop {
        let current_player = if move_no % 2 == 0 { 1 } else { 2 };
        let possible_moves = get_possible_moves(chosen_game);
        let number_of_possible_moves = possible_moves.len();

        if number_of_possible_moves == 0 {
            println!("Game over!");
            return None;
        }

        println!("Player {}'s turn:", current_player);
        println!("Possible moves:");
        for (i, mov_str) in possible_moves.into_iter().enumerate() {
            println!("{}: {}", i + 1, mov_str);
        }
        print!("Enter a move [1 to {}]: ", number_of_possible_moves);
        io::stdout().flush().unwrap();

        let mut line = String::new();
        io::stdin().read_line(&mut line).unwrap();
        let mov = match line.as_str().trim().parse::<usize>() {
            Ok(entered_move) => entered_move - 1,
            Err(_) => {
                println!("Invalid move!");
                continue;
            }
        };
        if mov >= number_of_possible_moves {
            println!("Invalid move!");
            continue;
        }

        return match chosen_game {
            ChosenGame::FoxAndHounds(game) => Some(ChosenGame::FoxAndHounds(
                game.execute_move(game.possible_moves()[mov]),
            )),
            ChosenGame::NoughtsAndCrosses(game) => Some(ChosenGame::NoughtsAndCrosses(
                game.execute_move(game.possible_moves()[mov]),
            )),
        };
    }
}

fn print_game(chosen_game: &ChosenGame) {
    match chosen_game {
        ChosenGame::FoxAndHounds(game) => {
            println!("{}\n", game);
        }
        ChosenGame::NoughtsAndCrosses(game) => {
            println!("{}\n", game);
        }
    }
}

fn main() {
    let mut game = choose_game();
    let game_type = choose_game_type();
    let mut move_no = 0;

    loop {
        print_game(&game);

        let next_game = match move_type_for_move(&game_type, move_no) {
            MoveType::PlayerMove => do_human_move(&game, move_no),
            MoveType::AIMove => do_ai_move(&game),
        };

        game = match next_game {
            Some(next_game) => next_game,
            None => {
                println!("Game over!");
                return;
            }
        };

        move_no += 1;
    }
}
